/* This file is part of rpipe
   Copyright (C) 2018 Sergey Poznyakoff
   License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
   This is free software: you are free to change and redistribute it.
   There is NO WARRANTY, to the extent permitted by law.
*/
#include <stdlib.h>

void error(int status, int err, char const *fmt, ...);
void *xmalloc(size_t s);
char *xstrdup(char const *s);
void *xcalloc(size_t nmemb, size_t size);
void *xrealloc(void *p, size_t size);
void *x2nrealloc(void *p, size_t *pn, size_t s);


#define IO_BUF_SIZE 512

/* Circular queue */
struct cqueue {
	char buf[IO_BUF_SIZE+1];   /* Cyclic buffer */
	ssize_t head;              /* Buffer head (read end) */
	ssize_t tail;              /* Buffer tail (write end) */
};

void cqueue_init(struct cqueue *cq);
size_t cqueue_size(struct cqueue *cq);

static inline size_t
cqueue_max(struct cqueue *cq)
{
	return sizeof(cq->buf);
}

static inline int
cqueue_is_full(struct cqueue *cq)
{
	return cqueue_size(cq) == cqueue_max(cq) - 1;
}

int cqueue_enqueue(struct cqueue *cq, int c);
int cqueue_dequeue(struct cqueue *cq);

struct sockaddr;
char *sockaddr_str(struct sockaddr *sa, int salen);
struct addrinfo;
struct addrinfo *net_address_parse(char const *addrstr, int server);

struct runas {
	uid_t uid;
	gid_t gid;
	gid_t *gidv;
	size_t gidc;
	size_t gidn;
};

typedef struct runas runas_t;

runas_t *runas_parse(char const *user_name, char const *group_names);
void runas(runas_t *r);


enum rpipe_server_type {
	rpipe_type_available,
	rpipe_type_listener, /* Listens for incoming requests */
	rpipe_type_init,
        /* Initializes the accepted connection, reads the first line,
	   starts the local command with the constructed command line */
	rpipe_type_net,
        /* Handles input/output with the net side of the pipe. */
	rpipe_type_prog,
	/* Output to the program */
};

struct rpipe_server_common {    
	enum rpipe_server_type type; /* Server type */
	char *str;                /* String representation */
	struct pollfd *pollfd;    /* File descriptor and configured events */
	union rpipe_server *next; /* Link to the next server */
};

struct rpipe_server_listener {    /* Listener server */
	/* Common part */
	enum rpipe_server_type type; /* Server type */
	char *str;                /* String representation */
	struct pollfd *pollfd;    /* File descriptor and configured events */
	union rpipe_server *next; /* Link to the next server */
	/* Server-specific part */
	int addrlen;              /* Length of the address */
	struct sockaddr *addr;    /* Address listened to */
};

struct rpipe_server_net {         /* Net I/O server. */
	/* Common part */
	enum rpipe_server_type type; /* Server type: rpipe_type_init or
				        rpipe_type_net */
	char *str;                /* String representation */
	struct pollfd *pollfd;    /* File descriptor and configured events */
	union rpipe_server *next; /* Link to the next server */
	/* Server-specific part */
	int addrlen;              /* Length of the address */
	struct sockaddr *addr;    /* Remote address */
	char buf[IO_BUF_SIZE];    /* Initial line buffer */
	size_t size;              /* Number of bytes in the buffer */
	size_t start;             /* Output position */
	struct rpipe_server_prog *prog; /* Pointer to the associated
					   prog server */
};

enum prog_state {
	prog_startup,
	prog_running,
	prog_terminating
};

struct rpipe_server_prog {        /* Output server */
	/* Common part */
	enum rpipe_server_type type; /* Server type */
	char *str;                /* String representation */
	struct pollfd *pollfd;    /* File descriptor and configured events */
	union rpipe_server *next; /* Link to the next server */
	/* Server-specific part */
	pid_t pid;                /* PID */
	struct cqueue cq;         /* I/O queue */
	enum prog_state state;
	struct rpipe_server_net *net;
};

typedef union rpipe_server {
	struct rpipe_server_common comm;
	struct rpipe_server_listener listener;
	struct rpipe_server_net net;
	struct rpipe_server_prog prog;
} rpipe_server_t;

void signal_setup(void (*handler)(int sig), int *sigv, int sigc);

void rpipe_client(struct addrinfo *, int argc, char **argv);
void rpipe_server(struct addrinfo *, int np, int argc, char **argv);

extern int verbose;
extern runas_t *runas_privs;
